package org.greenstone.gsdl3.gs3build.classifier;

import java.io.*;

import java.util.List;
import java.util.ArrayList;
import java.util.Iterator;

import org.xml.sax.XMLReader;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;
import org.xml.sax.Attributes;
import org.xml.sax.helpers.XMLReaderFactory;
import org.xml.sax.helpers.DefaultHandler;

import org.greenstone.gsdl3.gs3build.doctypes.DocumentID;
import org.greenstone.gsdl3.gs3build.doctypes.DocumentInterface;
import org.greenstone.gsdl3.gs3build.util.XMLTools;
import org.greenstone.gsdl3.gs3build.util.GS3SQLConnection;

public class HierarchyClassifier implements ClassifierInterface
{ 
  class HierarchyClassifierObserver
  { List              classifications;
    DocumentInterface document;

    public HierarchyClassifierObserver(DocumentInterface document)
    { this.document = document;
    }

    public void recordClassification(String label)
    { this.document.addDocumentMetadata("gsdl3", "classified", label);
    }

    public List getClassifications()
    { return this.classifications;
    }
  }

  class HierarchyNode
  { String descriptor;    // the textual descriptor used on screen or long-hand
    String name;          // the index number, letter assignment or other item
                          //   used to identify the position of the item in the
                          //   hierarchy
    String id;            // an identifier used by the GLI for maintenance
                          //   purposes; this plays no active role in the
                          //   rebuilding process (at the moment)
    List childNodes;      // the child classification nodes of this node
    List childDocs;       // the child documents of this node
    List matches;         // the other metadata values that may be matched
                          //   against the classifier
    HierarchyNode parent; // the parent of the node

    /**
     *  Simple node
     */
    public HierarchyNode()
    { this.descriptor = null;
      this.name       = null;
      this.id         = null;
      this.childNodes = new ArrayList();
      this.childDocs  = new ArrayList();
      this.parent     = null;
    }
    
    public void addChild(HierarchyNode child)
    { this.childNodes.add(child);
      child.setParent(this);
    }

    public void addDocument(DocumentID document)
    { this.childDocs.add(document);
    }

    public void setParent(HierarchyNode parent)
    { this.parent = parent;
    }
    
    public HierarchyNode getParent()
    { return this.parent;
    }

    public void setDescriptor(String descriptor)
    { this.descriptor = descriptor;
    }

    public void setID(String id)
    { this.id = id;
    }

    public void setName(String name)
    { this.name = name;
    }

    public void addMatch(String match)
    { this.matches.add(match);
    }

    public boolean isMatch (String toMatch)
    { Iterator thisMatch = this.matches.iterator();
			
      while (thisMatch.hasNext())
      { String thisMatchText = thisMatch.next().toString();

        if (thisMatchText.equals(toMatch))
	{ return true;
	}
      }
      return false;
    }

    public void getClassifications(DocumentID documentID, List values, 
				   HierarchyClassifierObserver observer)
    { Iterator valueList = values.iterator();

      while (valueList.hasNext())
      {	if (this.isMatch(valueList.next().toString()))
	{ observer.recordClassification(this.name);
	  this.addDocument(documentID);
	}
      }
      Iterator childList = this.childNodes.iterator();

      while (childList.hasNext())
      { ((HierarchyNode) childList.next()).getClassifications(documentID, values, observer);
      }
    }
  }

  /**
   *  An inner class to handle Classifier files
   */
  class HierarchyHandler extends DefaultHandler
  { StringBuffer description;
    StringBuffer name;
    StringBuffer id;
    StringBuffer match;
    boolean inElement;
    HierarchyNode rootNode = null;
    HierarchyNode currentNode = null;

    HierarchyHandler()
    { super();
 
      this.name        = null;
      this.description = null;
      this.id          = null;
      this.match       = null;
    }

    public void startElement(String URI, String localName, String qName, Attributes attributes)
    { if (localName.equals("Classification"))
      {	// create a new node in the hierarchy
	HierarchyNode node = new HierarchyNode();

	// Initialise the tree
	if (currentNode == null)
	{ if (this.rootNode == null) // which it should be
	  { this.rootNode = node;
	  }
	  else // insert a new root node, and set this as the peer
	       // of the original root...
	  { HierarchyNode newRoot = new HierarchyNode();
	    newRoot.addChild(this.rootNode);
	    newRoot.addChild(node);
	    this.rootNode = newRoot;
	  }
	}
        // Add the node
	else
	{ this.currentNode.addChild(node);
	}
	this.currentNode = node;
      }
      else if (localName.equals("Name"))
      {	this.name = new StringBuffer();
      }
      else if (localName.equals("ID"))
      {	this.id   = new StringBuffer();
      }
      else if (localName.equals("Description"))
      {	this.description = new StringBuffer();
      }
      else if (localName.equals("Matches"))
      {	this.match = new StringBuffer();
      }
      else if (localName.equals("Children"))
      {	
      }
    }

    public void endElement(String URI, String localName, String qName)
    { if (localName.equals("Classification"))
      {	this.currentNode = this.currentNode.getParent();
      }
      else if (localName.equals("Name"))
      { this.currentNode.setName(XMLTools.cleanString(this.name.toString()));
        this.name = null;
      }
      else if (localName.equals("ID"))
      {	this.currentNode.setID(XMLTools.cleanString(this.id.toString()));
        this.id = null;
      }
      else if (localName.equals("Description"))
      { this.currentNode.setDescriptor(XMLTools.cleanString(this.description.toString()));
        this.description = null;
      }
      else if (localName.equals("Matches"))
      { this.currentNode.addMatch(XMLTools.cleanString(this.match.toString()));
        this.match = null;
      }
      else if (localName.equals("Children"))
      {	
      }
    }

    public void characters(char c[], int start, int length)
    { String string = new String(c, start, length);
      if (this.id != null)
      { this.id.append(string);
      }
      else if (this.name != null)
      { this.name.append(string);
      }
      else if (this.description != null)
      { this.description.append(string);
      }
      else if (this.match != null)
      { this.match.append(string);
      }
    }

    public HierarchyNode getHierarchy()
    { return this.rootNode;
    }
  }

  private HierarchyNode hierarchy;
  private List          fields;

  public HierarchyClassifier(File basefile, List fields)
  {	
    try 
    { XMLReader reader = XMLReaderFactory.createXMLReader();
      HierarchyHandler handler = new HierarchyHandler();
      reader.setContentHandler(handler);
      reader.setErrorHandler(handler);

      // A Classifier consists of one file only - get it.
      FileReader fileReader = new FileReader(basefile);
      reader.parse(new InputSource(fileReader));

      this.hierarchy = handler.getHierarchy();
      this.fields = fields;
    }
    catch (SAXException saxException)
    { // TODO: log error
    }
    catch (java.io.FileNotFoundException fileException)
    {
    }
    catch (java.io.IOException ioException)
    {
    }
  }

  public void configure(List parameters)
  {
  }

  public void setDatabase(GS3SQLConnection database)
  {
  }

  /**
   *  Classify a particular document - the document id will be used to record its
   *  membership of any classifier nodes, and the document interface will also be
   *  used to assign back knowledge of any classifiers to which the document is
   *  assigned.
   *
   *  @param <code>DocumentID</code> the document identifier
   *  @param <code>DocumentInterface</code> the document interface
   */
  public boolean classifyDocument(DocumentID documentID, DocumentInterface document)
  { // the observer records the assignment of documents to classifications
    HierarchyClassifierObserver classifyObserver = new HierarchyClassifierObserver(document);
		
    // get the metadata item from the document
    Iterator thisField = this.fields.iterator();

    // for every field listed as being a scannable field...
    while (thisField.hasNext())
    { String fieldName = thisField.next().toString();
			
      // ...get the values for that field...
      List values = document.getDocumentMetadataItem(fieldName);

      if (values == null)
      {	continue;
      }

      // ...and send them to the classifier
      this.hierarchy.getClassifications(documentID, values, classifyObserver);
    }
    return true;
  }

  public int getClassifierPasses()
  { return 1;
  }

  public void startClassifierPass(int pass)
  {
  }
  
  public void endClassifierPass(int pass)
  {
  }
  
  public void completeClassification()
  { // TODO: store the classifications to file...
  }
}
